# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL
# Copyright (c) 2025 Open Brain Institute

import os

import numpy as np
import pandas as pd
import pytest
import re
from bluepysnap import Circuit
from numpy.testing import assert_array_equal, assert_allclose

from utils import TEST_DATA_DIR

import connectome_manipulator.connectome_comparison.connectivity as test_module


def _get_conn(edges_table, src_ids, tgt_ids, nodes, group_by):
    """Extract connectivity matrices from edges table"""
    conns, cnts = np.unique(
        edges_table[["@source_node", "@target_node"]], axis=0, return_counts=True
    )
    adj_mat = np.zeros((len(src_ids), len(tgt_ids)), dtype=bool)
    cnt_mat = np.zeros((len(src_ids), len(tgt_ids)), dtype=float) * np.nan
    for (_s, _t), _c in zip(conns, cnts):
        adj_mat[np.where(src_ids == _s)[0], np.where(tgt_ids == _t)[0]] = True
        cnt_mat[np.where(src_ids == _s)[0], np.where(tgt_ids == _t)[0]] = _c

    if not isinstance(group_by, tuple):
        group_by = (group_by, group_by)
    if group_by[0] is None:
        src_grp = np.zeros_like(src_ids)
    else:
        src_grp = nodes[0].get(src_ids, properties=group_by[0]).values
    if group_by[1] is None:
        tgt_grp = np.zeros_like(tgt_ids)
    else:
        tgt_grp = nodes[1].get(tgt_ids, properties=group_by[1]).values

    df_prob = pd.DataFrame(adj_mat, columns=tgt_grp)
    df_prob["src"] = src_grp
    df_prob = df_prob.melt("src", var_name="tgt", value_name="data")
    df_prob = df_prob.groupby(["src", "tgt"]).mean() * 100  # in %

    df_nsyn = pd.DataFrame(cnt_mat, columns=tgt_grp)
    df_nsyn["src"] = src_grp
    df_nsyn = df_nsyn.melt("src", var_name="tgt", value_name="data")

    df_nsyn_mean = df_nsyn.groupby(["src", "tgt"]).mean().fillna(0.0)
    df_nsyn_std = df_nsyn.groupby(["src", "tgt"]).std(ddof=0).fillna(0.0)
    df_nsyn_sem = df_nsyn.groupby(["src", "tgt"]).sem(ddof=0).fillna(0.0)
    df_nsyn_min = df_nsyn.groupby(["src", "tgt"]).min().fillna(0.0)
    df_nsyn_max = df_nsyn.groupby(["src", "tgt"]).max().fillna(0.0)

    return (
        df_prob.unstack(),
        df_nsyn_mean.unstack(),
        df_nsyn_std.unstack(),
        df_nsyn_sem.unstack(),
        df_nsyn_min.unstack(),
        df_nsyn_max.unstack(),
    )


def _check_conn(
    res, df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max, group_by
):
    for _key in [
        "conn_prob",
        "nsyn_conn",
        "nsyn_conn_std",
        "nsyn_conn_sem",
        "nsyn_conn_min",
        "nsyn_conn_max",
        "common",
    ]:
        assert _key in res, f'ERROR: Results key "{_key}" missing!'
    for _key in ["src_group_values", "tgt_group_values"]:
        assert _key in res["common"], f'ERROR: Results key "{_key}" in "common" missing!'
    for _key in [
        "conn_prob",
        "nsyn_conn",
        "nsyn_conn_std",
        "nsyn_conn_sem",
        "nsyn_conn_min",
        "nsyn_conn_max",
    ]:
        assert "data" in res[_key], f'ERROR: Results key "data" in "{_key}" missing!'

    if not isinstance(group_by, tuple):
        group_by = (group_by, group_by)
    if group_by[0] is None:
        assert_array_equal(
            res["common"]["src_group_values"], [None], "ERROR: Source group mismatch!"
        )
    else:
        assert_array_equal(
            res["common"]["src_group_values"],
            df_prob.index.to_numpy(),
            "ERROR: Source group mismatch!",
        )
    if group_by[1] is None:
        assert_array_equal(
            res["common"]["tgt_group_values"], [None], "ERROR: Target group mismatch!"
        )
    else:
        assert_array_equal(
            res["common"]["tgt_group_values"],
            df_prob.columns.get_level_values(1).to_numpy(),
            "ERROR: Target group mismatch!",
        )
    assert_allclose(
        res["conn_prob"]["data"], df_prob.to_numpy(), err_msg="ERROR: Conn. prob. mismatch!"
    )
    assert_allclose(
        res["nsyn_conn"]["data"],
        df_nsyn_mean.to_numpy(),
        err_msg="ERROR: Nsyn/conn (mean) mismatch!",
    )
    assert_allclose(
        res["nsyn_conn_std"]["data"],
        df_nsyn_std.to_numpy(),
        err_msg="ERROR: Nsyn/conn (std) mismatch!",
    )
    assert_allclose(
        res["nsyn_conn_sem"]["data"],
        df_nsyn_sem.to_numpy(),
        err_msg="ERROR: Nsyn/conn (sem) mismatch!",
    )
    assert_allclose(
        res["nsyn_conn_min"]["data"],
        df_nsyn_min.to_numpy(),
        err_msg="ERROR: Nsyn/conn (min) mismatch!",
    )
    assert_allclose(
        res["nsyn_conn_max"]["data"],
        df_nsyn_max.to_numpy(),
        err_msg="ERROR: Nsyn/conn (max) mismatch!",
    )


def _check_empty(res, src_ids, tgt_ids, nodes, group_by, skip):
    nsrc = len(nodes[0].property_values(group_by))
    ntgt = len(nodes[1].property_values(group_by))
    res_keys = [
        "nsyn_conn",
        "nsyn_conn_std",
        "nsyn_conn_sem",
        "nsyn_conn_min",
        "nsyn_conn_max",
        "conn_prob",
    ]
    for key in res_keys:
        if len(src_ids) == 0:  # No src selection
            if skip:
                assert res[key]["data"].shape[0] == 0
            else:
                assert res[key]["data"].shape[0] == nsrc
        else:
            assert res[key]["data"].shape[0] > 0

        if len(tgt_ids) == 0:  # No tgt selection
            if skip:
                assert res[key]["data"].shape[1] == 0
            else:
                assert res[key]["data"].shape[1] == ntgt
        else:
            assert res[key]["data"].shape[1] > 0


def test_connectivity():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]
    edges_table = edges.afferent_edges(nodes[1].ids(), properties=edges.property_names)

    # Case 1: Invalid inputs
    ## (a) Invalid population name
    popul_name = "INVALID_POPULATION_NAME"
    with pytest.raises(
        AssertionError, match=re.escape(f'Population "{popul_name}" not found in edges file')
    ):
        res = test_module.compute(
            circuit,
            sel_src=None,
            sel_dest=None,
            group_by=None,
            skip_empty_groups=False,
            edges_popul_name=popul_name,
        )

    ## (b) Invalid group-by
    popul_name = "nodeA__nodeA__chemical"
    with pytest.raises(
        AssertionError,
        match=re.escape("'group_by' must be a tuple with two elements for source/target neurons"),
    ):
        res = test_module.compute(
            circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=("mtype",)
        )
    with pytest.raises(
        AssertionError, match=re.escape("Source 'group_by' must be a string (or None)")
    ):
        res = test_module.compute(
            circuit,
            sel_src=None,
            sel_dest=None,
            edges_popul_name=popul_name,
            group_by=(123, "mtype"),
        )
    with pytest.raises(
        AssertionError, match=re.escape("Target 'group_by' must be a string (or None)")
    ):
        res = test_module.compute(
            circuit,
            sel_src=None,
            sel_dest=None,
            edges_popul_name=popul_name,
            group_by=("mtype", 123),
        )

    # Case 2: Full circuit
    ## (a) W/o group-by
    group_by = None
    res = test_module.compute(
        circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=group_by
    )
    df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max = _get_conn(
        edges_table, nodes[0].ids(), nodes[1].ids(), nodes, group_by
    )
    _check_conn(
        res, df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max, group_by
    )

    ## (b) W/ group-by
    group_by = "layer"
    res = test_module.compute(
        circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=group_by
    )
    df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max = _get_conn(
        edges_table, nodes[0].ids(), nodes[1].ids(), nodes, group_by
    )
    _check_conn(
        res, df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max, group_by
    )

    ## (c) W/ group-by (src only)
    group_by = ("layer", None)
    res = test_module.compute(
        circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=group_by
    )
    df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max = _get_conn(
        edges_table, nodes[0].ids(), nodes[1].ids(), nodes, group_by
    )
    _check_conn(
        res, df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max, group_by
    )

    ## (d) W/ group-by (tgt only)
    group_by = (None, "layer")
    res = test_module.compute(
        circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=group_by
    )
    df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max = _get_conn(
        edges_table, nodes[0].ids(), nodes[1].ids(), nodes, group_by
    )
    _check_conn(
        res, df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max, group_by
    )

    ## (e) W/ group-by (different src/tgt)
    group_by = ("layer", "mtype")
    res = test_module.compute(
        circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=group_by
    )
    df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max = _get_conn(
        edges_table, nodes[0].ids(), nodes[1].ids(), nodes, group_by
    )
    _check_conn(
        res, df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max, group_by
    )

    # Case 3: Partial circuit (layer by layer) w/ group-by
    group_by = "mtype"
    for _src_lay in nodes[0].property_values("layer"):
        for _tgt_lay in nodes[1].property_values("layer"):
            sel_src = {"layer": _src_lay}
            sel_tgt = {"layer": _tgt_lay}
            res = test_module.compute(
                circuit,
                sel_src=sel_src,
                sel_dest=sel_tgt,
                edges_popul_name=popul_name,
                group_by=group_by,
                skip_empty_groups=True,
            )
            df_prob, df_nsyn_mean, df_nsyn_std, df_nsyn_sem, df_nsyn_min, df_nsyn_max = _get_conn(
                edges_table, nodes[0].ids(sel_src), nodes[1].ids(sel_tgt), nodes, group_by
            )
            _check_conn(
                res,
                df_prob,
                df_nsyn_mean,
                df_nsyn_std,
                df_nsyn_sem,
                df_nsyn_min,
                df_nsyn_max,
                group_by,
            )

    # Case 4: Empty source/target node selection (with and w/o skip)
    group_by = "layer"
    for skip in [True, False]:
        ## (a) src sel empty
        sel_src = {"layer": "UNKNOWN"}
        sel_tgt = None
        res = test_module.compute(
            circuit,
            sel_src=sel_src,
            sel_dest=sel_tgt,
            edges_popul_name=popul_name,
            group_by=group_by,
            skip_empty_groups=skip,
        )
        _check_empty(res, nodes[0].ids(sel_src), nodes[1].ids(sel_tgt), nodes, group_by, skip)

        ## (b) tgt sel empty
        sel_src = None
        sel_tgt = {"layer": "UNKNOWN"}
        res = test_module.compute(
            circuit,
            sel_src=sel_src,
            sel_dest=sel_tgt,
            edges_popul_name=popul_name,
            group_by=group_by,
            skip_empty_groups=skip,
        )
        _check_empty(res, nodes[0].ids(sel_src), nodes[1].ids(sel_tgt), nodes, group_by, skip)

        ## (c) src & tgt sel empty
        sel_src = {"layer": "UNKNOWN"}
        sel_tgt = {"layer": "UNKNOWN"}
        res = test_module.compute(
            circuit,
            sel_src=sel_src,
            sel_dest=sel_tgt,
            edges_popul_name=popul_name,
            group_by=group_by,
            skip_empty_groups=skip,
        )
        _check_empty(res, nodes[0].ids(sel_src), nodes[1].ids(sel_tgt), nodes, group_by, skip)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL
# Copyright (c) 2025 Open Brain Institute

import os
import re
import json
import pytest
from pathlib import Path
from cached_property import cached_property
from connectome_manipulator import utils as test_module
from bluepysnap.circuit import Circuit
from bluepysnap.nodes import NodePopulation
from unittest.mock import MagicMock

from utils import TEST_DATA_DIR


@pytest.mark.parametrize(
    "base_dir, path, expected",
    [
        ("/a/b", "$c", "$c"),
        ("/a/b", "", ""),
        ("/a/b", "/c/d", "/c/d"),
        ("/a/b", "/a/b/c", "$BASE_DIR/c"),
        ("/a/b", "/a/b/c/d", "$BASE_DIR/c/d"),
        ("/a/b/c", "/a/b/c/d", "$BASE_DIR/d"),
        ("/a/b", "./c", "$BASE_DIR/c"),
        ("/a/b", "c/d", "$BASE_DIR/c/d"),
    ],
)
def tested_reduce_path(base_dir, path, expected):
    result = test_module._reduce_path(path, Path(base_dir))
    assert result == expected


def test_reduce_config_paths__raises_relative_config_dir():
    expected = re.escape("Circuit config's directory is not absolute: .")
    with pytest.raises(ValueError, match=expected):
        test_module.reduce_config_paths(None, ".")


def test_reduce_config_paths__not_resolved_config():
    config_path = Path(TEST_DATA_DIR, "circuit_sonata.json")
    config = json.loads(config_path.read_bytes())
    contains = "A reduced config with absolute paths and no manifest must be provided."
    with pytest.raises(ValueError, match=contains):
        test_module.reduce_config_paths(config, "/")


def test_reduce_config_paths():
    config_path = Path(TEST_DATA_DIR, "circuit_sonata.json")

    reduced_config = Circuit(config_path).config

    res = test_module.reduce_config_paths(reduced_config, TEST_DATA_DIR)

    assert res["version"] == 2
    assert res["node_sets_file"] == "$BASE_DIR/node_sets.json"
    assert res["networks"] == {
        "nodes": [
            {
                "nodes_file": "$BASE_DIR/nodes.h5",
                "populations": {
                    "nodeA": {
                        "type": "biophysical",
                        "morphologies_dir": "$BASE_DIR/swc",
                        "biophysical_neuron_models_dir": "$BASE_DIR",
                    },
                },
            },
        ],
        "edges": [
            {
                "edges_file": "$BASE_DIR/edges.h5",
                "populations": {"nodeA__nodeA__chemical": {"type": "chemical"}},
            },
        ],
    }


def test_check_grouping():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    src_nodes = nodes[0]
    tgt_nodes = nodes[1]

    # Create mock source/target nodes without "layer" property
    src_nodes_mock = MagicMock()
    src_nodes_mock.property_names = {_p for _p in src_nodes.property_names if _p != "layer"}
    src_nodes_mock.get.return_value = src_nodes.get().drop("layer", axis=1)

    tgt_nodes_mock = MagicMock()
    tgt_nodes_mock.property_names = {_p for _p in tgt_nodes.property_names if _p != "layer"}
    tgt_nodes_mock.get.return_value = tgt_nodes.get().drop("layer", axis=1)

    # Case 1: No node populations provided
    ## (a) No grouping --> OK
    group_by = None
    res = test_module.check_grouping(group_by, src_nodes=None, tgt_nodes=None)
    assert res == (None, None)

    ## (b) No grouping (tuple) --> OK
    group_by = (None, None)
    res = test_module.check_grouping(group_by, src_nodes=None, tgt_nodes=None)
    assert res == (None, None)

    ## (c) No grouping (wrong tuple) --> ERROR
    group_by = (None, None, None)
    with pytest.raises(
        AssertionError,
        match=re.escape("'group_by' must be a tuple with two elements for source/target neurons"),
    ):
        res = test_module.check_grouping(group_by, src_nodes=None, tgt_nodes=None)

    ## (d) Grouping (str) --> OK
    group_by = "any"
    res = test_module.check_grouping(group_by, src_nodes=None, tgt_nodes=None)
    assert res == ("any", "any")

    ## (e) Grouping (tuple) --> OK
    group_by = ("any-A", "any-B")
    res = test_module.check_grouping(group_by, src_nodes=None, tgt_nodes=None)
    assert res == ("any-A", "any-B")

    # Case 2: Same source/target node populations provided
    ## (a) No grouping --> OK
    group_by = None
    res = test_module.check_grouping(group_by, src_nodes=src_nodes, tgt_nodes=tgt_nodes)
    assert res == (None, None)

    ## (b) No grouping (tuple) --> OK
    group_by = (None, None)
    res = test_module.check_grouping(group_by, src_nodes=src_nodes, tgt_nodes=tgt_nodes)
    assert res == (None, None)

    ## (c) Invalid grouping (str) --> ERROR
    group_by = "invalid"
    with pytest.raises(
        AssertionError,
        match=re.escape(
            f"Grouping property '{group_by}' does not exist in either source or target node population"
        ),
    ):
        res = test_module.check_grouping(group_by, src_nodes=src_nodes, tgt_nodes=tgt_nodes)

    ## (d) Invalid source grouping (str) --> ERROR
    group_by = ("invalid", None)
    with pytest.raises(
        AssertionError,
        match=re.escape(
            f"Grouping property '{group_by[0]}' does not exist in source node population"
        ),
    ):
        res = test_module.check_grouping(group_by, src_nodes=src_nodes, tgt_nodes=tgt_nodes)

    ## (e) Invalid target grouping (str) --> ERROR
    group_by = (None, "invalid")
    with pytest.raises(
        AssertionError,
        match=re.escape(
            f"Grouping property '{group_by[1]}' does not exist in target node population"
        ),
    ):
        res = test_module.check_grouping(group_by, src_nodes=src_nodes, tgt_nodes=tgt_nodes)

    ## (f) Valid grouping (str) --> OK
    group_by = "layer"
    res = test_module.check_grouping(group_by, src_nodes=src_nodes, tgt_nodes=tgt_nodes)
    assert res == ("layer", "layer")

    ## (g) Valid grouping (tuple) --> OK
    group_by = ("layer", "mtype")
    res = test_module.check_grouping(group_by, src_nodes=src_nodes, tgt_nodes=tgt_nodes)
    assert res == ("layer", "mtype")

    # Case 3: Different source/target node populations provided
    ## (a) Invalid grouping in src/tgt (str) --> ERROR
    group_by = "invalid"
    with pytest.raises(
        AssertionError,
        match=re.escape(
            f"Grouping property '{group_by}' does not exist in either source or target node population"
        ),
    ):
        res = test_module.check_grouping(group_by, src_nodes=src_nodes, tgt_nodes=tgt_nodes_mock)

    ## (b) Valid grouping in src only (str) --> OK (will ignore tgt)
    group_by = "layer"
    res = test_module.check_grouping(group_by, src_nodes=src_nodes, tgt_nodes=tgt_nodes_mock)
    assert res == ("layer", None)

    ## (c) Valid grouping in src only (tuple) --> ERROR (will not ignore tgt)
    group_by = ("layer", "layer")
    with pytest.raises(
        AssertionError,
        match=re.escape(
            f"Grouping property '{group_by[1]}' does not exist in target node population"
        ),
    ):
        res = test_module.check_grouping(group_by, src_nodes=src_nodes, tgt_nodes=tgt_nodes_mock)

    ## (d) Valid grouping in tgt only (str) --> OK (will ignore src)
    group_by = "layer"
    res = test_module.check_grouping(group_by, src_nodes=src_nodes_mock, tgt_nodes=tgt_nodes)
    assert res == (None, "layer")

    ## (e) Valid grouping in tgt only (tuple) --> ERROR (will not ignore src)
    group_by = ("layer", "layer")
    with pytest.raises(
        AssertionError,
        match=re.escape(
            f"Grouping property '{group_by[1]}' does not exist in source node population"
        ),
    ):
        res = test_module.check_grouping(group_by, src_nodes=src_nodes_mock, tgt_nodes=tgt_nodes)

    ## (f) Valid grouping in src/tgt (str) --> OK
    group_by = "mtype"
    res = test_module.check_grouping(group_by, src_nodes=src_nodes_mock, tgt_nodes=tgt_nodes)
    assert res == ("mtype", "mtype")

    ## (g) Valid grouping in src/tgt (tuple) --> OK
    group_by = ("mtype", "layer")
    res = test_module.check_grouping(group_by, src_nodes=src_nodes_mock, tgt_nodes=tgt_nodes)
    assert res == ("mtype", "layer")



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL
# Copyright (c) 2025 Open Brain Institute

import os

import numpy as np
import pandas as pd
import pytest
import re
from bluepysnap import Circuit
from numpy.testing import assert_array_equal, assert_allclose

from utils import TEST_DATA_DIR

import connectome_manipulator.connectome_comparison.properties as test_module


def _get_props(edges_table, src_ids, tgt_ids, nodes, group_by, prop_name):
    """Extract connectivity matrices from edges table"""
    conns = np.unique(edges_table[["@source_node", "@target_node"]], axis=0)
    prop_mat_sum = np.zeros((len(src_ids), len(tgt_ids)), dtype=float) * np.nan
    prop_mat_cnt = np.zeros((len(src_ids), len(tgt_ids)), dtype=int)
    for _s, _t in conns:
        edge_sel = np.logical_and(
            edges_table["@source_node"] == _s, edges_table["@target_node"] == _t
        )
        prop_mat_sum[np.where(src_ids == _s)[0], np.where(tgt_ids == _t)[0]] = edges_table.loc[
            edge_sel, prop_name
        ].sum()
        prop_mat_cnt[np.where(src_ids == _s)[0], np.where(tgt_ids == _t)[0]] = np.sum(edge_sel)

    if not isinstance(group_by, tuple):
        group_by = (group_by, group_by)
    if group_by[0] is None:
        src_grp = np.zeros_like(src_ids)
    else:
        src_grp = nodes[0].get(src_ids, properties=group_by[0]).values
    if group_by[1] is None:
        tgt_grp = np.zeros_like(tgt_ids)
    else:
        tgt_grp = nodes[1].get(tgt_ids, properties=group_by[1]).values

    df_props_sum = pd.DataFrame(prop_mat_sum, columns=tgt_grp)
    df_props_sum["src"] = src_grp
    df_props_sum = df_props_sum.melt("src", var_name="tgt", value_name="data")
    df_props_sum = df_props_sum.groupby(["src", "tgt"]).sum()

    df_props_cnt = pd.DataFrame(prop_mat_cnt, columns=tgt_grp)
    df_props_cnt["src"] = src_grp
    df_props_cnt = df_props_cnt.melt("src", var_name="tgt", value_name="data")
    df_props_cnt = df_props_cnt.groupby(["src", "tgt"]).sum()

    df_props = df_props_sum / df_props_cnt

    return df_props.unstack()


def _check_props(res, df_props, group_by, prop_name):
    for _key in [prop_name, "common"]:
        assert _key in res, f'ERROR: Results key "{_key}" missing!'
    for _key in ["src_group_values", "tgt_group_values"]:
        assert _key in res["common"], f'ERROR: Results key "{_key}" in "common" missing!'
    assert "data" in res[prop_name], f'ERROR: Results key "data" in "{prop_name}" missing!'

    if not isinstance(group_by, tuple):
        group_by = (group_by, group_by)
    if group_by[0] is None:
        assert_array_equal(
            res["common"]["src_group_values"], [None], "ERROR: Source group mismatch!"
        )
    else:
        assert_array_equal(
            res["common"]["src_group_values"],
            df_props.index.to_numpy(),
            "ERROR: Source group mismatch!",
        )
    if group_by[1] is None:
        assert_array_equal(
            res["common"]["tgt_group_values"], [None], "ERROR: Target group mismatch!"
        )
    else:
        assert_array_equal(
            res["common"]["tgt_group_values"],
            df_props.columns.get_level_values(1).to_numpy(),
            "ERROR: Target group mismatch!",
        )
    assert_allclose(
        res[prop_name]["data"],
        df_props.to_numpy(),
        rtol=1e-06,
        err_msg=f'ERROR: "{prop_name}" mismatch!',
    )


def _check_empty(res, src_ids, tgt_ids, nodes, edges, group_by, skip):
    nsrc = len(nodes[0].property_values(group_by))
    ntgt = len(nodes[1].property_values(group_by))
    res_keys = list(edges.property_names)
    for key in res_keys:
        if len(src_ids) == 0:  # No src selection
            if skip:
                assert res[key]["data"].shape[0] == 0
            else:
                assert res[key]["data"].shape[0] == nsrc
        else:
            assert res[key]["data"].shape[0] > 0

        if len(tgt_ids) == 0:  # No tgt selection
            if skip:
                assert res[key]["data"].shape[1] == 0
            else:
                assert res[key]["data"].shape[1] == ntgt
        else:
            assert res[key]["data"].shape[1] > 0


def test_properties():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]
    edges_table = edges.afferent_edges(nodes[1].ids(), properties=edges.property_names)

    # Case 1: Invalid inputs
    ## (a) Invalid population name
    popul_name = "INVALID_POPULATION_NAME"
    with pytest.raises(
        AssertionError, match=re.escape(f'Population "{popul_name}" not found in edges file')
    ):
        res = test_module.compute(
            circuit,
            sel_src=None,
            sel_dest=None,
            group_by=None,
            skip_empty_groups=False,
            edges_popul_name=popul_name,
        )

    ## (b) Invalid group-by
    popul_name = "nodeA__nodeA__chemical"
    with pytest.raises(
        AssertionError,
        match=re.escape("'group_by' must be a tuple with two elements for source/target neurons"),
    ):
        res = test_module.compute(
            circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=("mtype",)
        )
    with pytest.raises(
        AssertionError, match=re.escape("Source 'group_by' must be a string (or None)")
    ):
        res = test_module.compute(
            circuit,
            sel_src=None,
            sel_dest=None,
            edges_popul_name=popul_name,
            group_by=(123, "mtype"),
        )
    with pytest.raises(
        AssertionError, match=re.escape("Target 'group_by' must be a string (or None)")
    ):
        res = test_module.compute(
            circuit,
            sel_src=None,
            sel_dest=None,
            edges_popul_name=popul_name,
            group_by=("mtype", 123),
        )

    # Case 2: Full circuit
    ## (a) W/o group-by
    group_by = None
    res = test_module.compute(
        circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=group_by
    )
    for eprop in edges.property_names:
        df_props = _get_props(edges_table, nodes[0].ids(), nodes[1].ids(), nodes, group_by, eprop)
        _check_props(res, df_props, group_by, eprop)

    ## (b) W/ group-by
    group_by = "layer"
    res = test_module.compute(
        circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=group_by
    )
    for eprop in edges.property_names:
        df_props = _get_props(edges_table, nodes[0].ids(), nodes[1].ids(), nodes, group_by, eprop)
        _check_props(res, df_props, group_by, eprop)

    ## (c) W/ group-by (src only)
    group_by = ("layer", None)
    res = test_module.compute(
        circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=group_by
    )
    for eprop in edges.property_names:
        df_props = _get_props(edges_table, nodes[0].ids(), nodes[1].ids(), nodes, group_by, eprop)
        _check_props(res, df_props, group_by, eprop)

    ## (d) W/ group-by (tgt only)
    group_by = (None, "layer")
    res = test_module.compute(
        circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=group_by
    )
    for eprop in edges.property_names:
        df_props = _get_props(edges_table, nodes[0].ids(), nodes[1].ids(), nodes, group_by, eprop)
        _check_props(res, df_props, group_by, eprop)

    ## (e) W/ group-by (different src/tgt)
    group_by = ("layer", "mtype")
    res = test_module.compute(
        circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name, group_by=group_by
    )
    for eprop in edges.property_names:
        df_props = _get_props(edges_table, nodes[0].ids(), nodes[1].ids(), nodes, group_by, eprop)
        _check_props(res, df_props, group_by, eprop)

    # Case 3: Partial circuit (layer by layer) w/ group-by
    group_by = "mtype"
    for _src_lay in nodes[0].property_values("layer"):
        for _tgt_lay in nodes[1].property_values("layer"):
            sel_src = {"layer": _src_lay}
            sel_tgt = {"layer": _tgt_lay}
            res = test_module.compute(
                circuit,
                sel_src=sel_src,
                sel_dest=sel_tgt,
                edges_popul_name=popul_name,
                group_by=group_by,
                skip_empty_groups=True,
            )
            for eprop in edges.property_names:
                df_props = _get_props(
                    edges_table,
                    nodes[0].ids(sel_src),
                    nodes[1].ids(sel_tgt),
                    nodes,
                    group_by,
                    eprop,
                )
                _check_props(res, df_props, group_by, eprop)

    # Case 4: Empty source/target node selection (with and w/o skip)
    group_by = "layer"
    for skip in [True, False]:
        ## (a) src sel empty
        sel_src = {"layer": "UNKNOWN"}
        sel_tgt = None
        res = test_module.compute(
            circuit,
            sel_src=sel_src,
            sel_dest=sel_tgt,
            edges_popul_name=popul_name,
            group_by=group_by,
            skip_empty_groups=skip,
        )
        _check_empty(
            res, nodes[0].ids(sel_src), nodes[1].ids(sel_tgt), nodes, edges, group_by, skip
        )

        ## (b) tgt sel empty
        sel_src = None
        sel_tgt = {"layer": "UNKNOWN"}
        res = test_module.compute(
            circuit,
            sel_src=sel_src,
            sel_dest=sel_tgt,
            edges_popul_name=popul_name,
            group_by=group_by,
            skip_empty_groups=skip,
        )
        _check_empty(
            res, nodes[0].ids(sel_src), nodes[1].ids(sel_tgt), nodes, edges, group_by, skip
        )

        ## (c) src & tgt sel empty
        sel_src = {"layer": "UNKNOWN"}
        sel_tgt = {"layer": "UNKNOWN"}
        res = test_module.compute(
            circuit,
            sel_src=sel_src,
            sel_dest=sel_tgt,
            edges_popul_name=popul_name,
            group_by=group_by,
            skip_empty_groups=skip,
        )
        _check_empty(
            res, nodes[0].ids(sel_src), nodes[1].ids(sel_tgt), nodes, edges, group_by, skip
        )



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL
# Copyright (c) 2025 Open Brain Institute

"""Various utility functions"""

import json
import os
from pathlib import Path
from typing import Any, Dict

import numpy as np


def _apply_recursively(func, obj, origin=(0, 0, 0)):
    origin = np.array(origin)

    if hasattr(obj, "soma"):
        obj.soma.points = origin + func(obj.soma.points - origin)
    for s in obj.iter():
        s.points = origin + func(s.points - origin)


def transform(obj, A):
    """Apply transformation matrix `A` to a given morphology object.

    Note:
        This comes out of bluepysnap

    Args:
        obj: Morphology / Section
        A: rotation matrix (4 x 4 NumPy array)
    """
    if A is None:
        return
    A = np.asarray(A)
    if A.shape != (4, 4):
        raise ValueError(f"`A` should be 4 x 4 matrix (got: {A.shape})")
    A = A.transpose()

    def func(p):
        return np.dot(np.column_stack((p, np.ones(len(p)))), A)[:, :3]

    _apply_recursively(func, obj)


def create_dir(path: os.PathLike) -> Path:
    """Create directory and parents if it doesn't already exist."""
    path = Path(path)
    path.mkdir(parents=True, exist_ok=True)
    return path


def load_json(filepath: os.PathLike) -> Dict[Any, Any]:
    """Load from JSON file."""
    return json.loads(Path(filepath).read_bytes())


def write_json(filepath: os.PathLike, data: Dict[Any, Any], indent: int = 2) -> None:
    """Write json file."""
    with open(filepath, "w", encoding="utf-8") as fd:
        json.dump(data, fd, indent=indent)


def reduce_config_paths(config: dict, config_dir: os.PathLike) -> dict:
    """Reduces absolute paths with respect to base_dir.

    Args:
        config (dict): Input config with absolute paths and no manifest.
        config_dir (str|Path): The path to the directory of the config.

    Returns:
        A new config with the paths reduced to being relative to the config dir.

    Note: Paths that are not in config_dir's tree will stay absolute.
    """
    if not Path(config_dir).is_absolute():
        raise ValueError(f"Circuit config's directory is not absolute: {config_dir}")

    if "manifest" in config:
        raise ValueError(
            f"A reduced config with absolute paths and no manifest must be provided.\n"
            f"Got instead: {config}"
        )

    reduced_config = {
        "version": config.get("version", "1"),
        "manifest": {"$BASE_DIR": "."},
    }

    if "node_sets_file" in config:
        reduced_config["node_sets_file"] = _reduce_path(config["node_sets_file"], config_dir)

    if "components" in config:
        # Reduce components dict recursively (may have nested dicts!)
        reduced_config["components"] = _reduce_dict(config["components"], config_dir)

    reducer = {
        "nodes_file": _reduce_path,
        "node_types_file": _reduce_path,
        "edges_file": _reduce_path,
        "edge_types_file": _reduce_path,
        "populations": _reduce_populations,
    }

    reduced_config["networks"] = {
        network_type: [
            {
                key: (reducer[key](value, config_dir) if key in reducer else value)
                for key, value in net_dict.items()
            }
            for net_dict in network_list
        ]
        for network_type, network_list in config["networks"].items()
    }
    return reduced_config


def _reduce_path(path: str, base_dir: os.PathLike) -> str:
    if not path or path.startswith("$"):
        return path

    path = Path(path)

    if path.is_absolute():
        if path.is_relative_to(base_dir):
            return str(Path("$BASE_DIR", path.relative_to(base_dir)))

    return str(Path("$BASE_DIR", path))


def _reduce_populations(populations_dict: dict, base_dir: os.PathLike) -> dict:
    def reduce_entry(key, value):
        if key.endswith(("file", "dir")):
            return _reduce_path(value, base_dir)

        if key == "alternate_morphologies":
            return {
                alt_key: _reduce_path(alt_path, base_dir) for alt_key, alt_path in value.items()
            }

        return value

    return {
        pop_name: {key: reduce_entry(key, value) for key, value in pop_dict.items()}
        for pop_name, pop_dict in populations_dict.items()
    }


def _reduce_dict(entry: dict, base_dir: os.PathLike) -> dict:
    reduced_dict = {}
    for key, value in entry.items():
        if isinstance(value, dict):
            reduced_dict[key] = _reduce_dict(value, base_dir)
        else:
            reduced_dict[key] = _reduce_path(value, base_dir)
    return reduced_dict


def _check_node_property(property_name, nodes):
    if nodes is None or property_name in nodes.property_names:
        return property_name
    else:
        return None


def check_grouping(group_by, src_nodes=None, tgt_nodes=None):
    """Helper function to check grouping for source/target neurons.

    Args:
        group_by (str/tuple): Neuron property name based on which to group connections, e.g., "synapse_class", "layer", or "mtype"; can be a tuple with two property names for source/target neurons or omitted
        src_nodes (bluepysnap.nodes.NodePopulation): Source node population (optional)
        tgt_nodes (bluepysnap.nodes.NodePopulation): Target node population (optional)

    Returns:
        str: Source grouping property name
        str: Target grouping property name

    Note:
        Includes some logic so that if a given property name only exists in either the source or target node
        population, it will apply grouping only on the population where it exists (instead of raising an error).
    """
    if not isinstance(group_by, tuple):
        if group_by is None:
            group_by = (group_by, group_by)
        else:
            group_by_lbl = str(group_by)
            # If only one grouping property is provided, propagate only to the populations where it actually exists
            group_by = (
                _check_node_property(group_by, src_nodes),
                _check_node_property(group_by, tgt_nodes),
            )
            # But make sure it exists in at least one of the source/target populations
            assert any(
                _g is not None for _g in group_by
            ), f"ERROR: Grouping property '{group_by_lbl}' does not exist in either source or target node population!"

    assert (
        len(group_by) == 2
    ), "ERROR: 'group_by' must be a tuple with two elements for source/target neurons!"

    src_group_by = group_by[0]
    tgt_group_by = group_by[1]

    assert src_group_by is None or isinstance(
        src_group_by, str
    ), "ERROR: Source 'group_by' must be a string (or None)!"
    assert tgt_group_by is None or isinstance(
        tgt_group_by, str
    ), "ERROR: Target 'group_by' must be a string (or None)!"

    if src_group_by is not None:
        assert (
            _check_node_property(src_group_by, src_nodes) is not None
        ), f"ERROR: Grouping property '{src_group_by}' does not exist in source node population!"
    if tgt_group_by is not None:
        assert (
            _check_node_property(tgt_group_by, tgt_nodes) is not None
        ), f"ERROR: Grouping property '{tgt_group_by}' does not exist in target node population!"

    return src_group_by, tgt_group_by


class ConsoleColors:
    """Helper class for formatting console text."""

    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, _, DEFAULT = range(30, 40)
    NORMAL, BOLD, DIM, UNDERLINED, BLINK, INVERTED, HIDDEN = [a << 8 for a in range(7)]

    # These are the sequences needed to control output
    _CHANGE_SEQ = "\033[{}m"
    _RESET_SEQ = "\033[0m"

    @classmethod
    def reset(cls):
        """Reset colors."""
        return cls._RESET_SEQ

    @classmethod
    def set_text_color(cls, color):
        """Change text color."""
        return cls._CHANGE_SEQ.format(color)

    @classmethod
    def format_text(cls, text, color, style=None):
        """Format the text."""
        style = (style or color) >> 8
        format_seq = str(color & 0x00FF) + ((";" + str(style)) if style else "")
        return cls._CHANGE_SEQ.format(format_seq) + text + cls._RESET_SEQ
